@@ -3,16 +3,18 @@ from __future__ import print_function
 from collections import Counter, defaultdict, deque
 from functools import partial, wraps
 from heapq import merge
-from itertools import chain, count, islice, repeat, takewhile, tee
+from itertools import chain, count, groupby, islice, repeat, takewhile, tee
 from operator import itemgetter
 from sys import version_info
 
-from six import string_types
+from six import binary_type, string_types, text_type
 from six.moves import filter, map, zip, zip_longest
 
 from .recipes import flatten, take
 
 __all__ = [
+    'adjacent',
+    'always_iterable',
     'bucket',
     'chunked',
     'collapse',
@@ -20,7 +22,9 @@ __all__ = [
     'consumer',
     'distinct_permutations',
     'distribute',
+    'divide',
     'first',
+    'groupby_transform',
     'ilen',
     'interleave_longest',
     'interleave',
@@ -42,7 +46,6 @@ __all__ = [
     'zip_offset',
 ]
 
-
 _marker = object()
 
 
@@ -95,7 +98,7 @@ def first(iterable, default=_marker):
 
 
 class peekable(object):
-    """Wrap an iterator to allow lookahead.
+    """Wrap an iterator to allow lookahead and prepending elements.
 
     Call ``peek()`` on the result to get the value that will next pop out of
     ``next()``, without advancing the iterator:
@@ -113,6 +116,21 @@ class peekable(object):
         >>> p.peek('hi')
         'hi'
 
+    peekables also offer a ``prepend()`` method which will insert items before
+    the remaining part of the underlying source iterator.
+
+        >>> p = peekable([1, 2, 3])
+        >>> p.prepend(10, 11, 12)
+        >>> next(p)
+        10
+        >>> p.peek()
+        11
+        >>> list(p)
+        [11, 12, 1, 2, 3]
+
+    Prepended items are treated by other peekable methods exactly as if they
+    had come from the source iterator.
+
     You may index the peekable to look ahead by more than one item.
     The values up to the index you specified will be cached.
     Index 0 is the item that will be returned by ``next()``, index 1 is the
@@ -125,8 +143,11 @@ class peekable(object):
         'b'
         >>> next(p)
         'a'
+        >>> p.prepend('x')
         >>> p[1]
-        'c'
+        'b'
+        >>> next(p)
+        'x'
         >>> next(p)
         'b'
 
@@ -135,10 +156,14 @@ class peekable(object):
     storage.
 
     To test whether there are more items in the iterator, examine the
-    peekable's truth value. If it is truthy, there are more items.
+    peekable's truth value. If it is truthy, there are more items (which may
+    have been prepended or obtained from the source iterator).
 
         >>> assert peekable([1])
-        >>> assert not peekable([])
+        >>> p = peekable([])
+        >>> assert not p
+        >>> p.prepend(1)
+        >>> assert p
 
     """
     def __init__(self, iterable):
@@ -175,6 +200,37 @@ class peekable(object):
                 return default
         return self._cache[0]
 
+    def prepend(self, *items):
+        """Stack up items to be the next ones returned from ``next()`` or
+        ``self.peek()``. The items will be returned in
+        first in, first out order::
+
+            >>> p = peekable([1, 2, 3])
+            >>> p.prepend(10, 11, 12)
+            >>> next(p)
+            10
+            >>> list(p)
+            [11, 12, 1, 2, 3]
+
+        It is possible, by prepending items, to "resurrect" a peekable that
+        previously raised ``StopIteration``.
+
+            >>> p = peekable([])
+            >>> next(p)
+            Traceback (most recent call last):
+              ...
+            StopIteration
+            >>> p.prepend(1)
+            >>> next(p)
+            1
+            >>> next(p)
+            Traceback (most recent call last):
+              ...
+            StopIteration
+
+        """
+        self._cache.extendleft(reversed(items))
+
     def __next__(self):
         if self._cache:
             return self._cache.popleft()
@@ -309,10 +365,11 @@ def ilen(iterable):
         >>> ilen(x for x in range(1000000) if x % 3 == 0)
         333334
 
-    This does, of course, consume the iterable, so handle it with care.
+    This consumes the iterable, so handle with care.
 
     """
-    return sum(1 for _ in iterable)
+    d = deque(enumerate(iterable, 1), maxlen=1)
+    return d[0][0] if d else 0
 
 
 def iterate(func, start):
@@ -630,24 +687,29 @@ def spy(iterable, n=1):
 
 def interleave(*iterables):
     """Return a new iterable yielding from each iterable in turn,
-    until the shortest is exhausted. Note that this is the same as
-    ``chain(*zip(*iterables))``.
+    until the shortest is exhausted.
 
         >>> list(interleave([1, 2, 3], [4, 5], [6, 7, 8]))
         [1, 4, 6, 2, 5, 7]
 
+    Note that this is the same as ``chain(*zip(*iterables))``.
+    For a version that doesn't terminate after the shortest iterable is
+    exhausted, see ``interleave_longest()``.
+
     """
     return chain.from_iterable(zip(*iterables))
 
 
 def interleave_longest(*iterables):
     """Return a new iterable yielding from each iterable in turn,
-    skipping any that are exhausted. Note that this is not the same as
-    ``chain(*zip_longest(*iterables))``.
+    skipping any that are exhausted.
 
         >>> list(interleave_longest([1, 2, 3], [4, 5], [6, 7, 8]))
         [1, 4, 6, 2, 5, 7, 3, 8]
 
+    Note that this is an alternate implementation of ``roundrobin()`` from the
+    itertools documentation.
+
     """
     i = chain.from_iterable(zip_longest(*iterables, fillvalue=_marker))
     return filter(lambda x: x is not _marker, i)
@@ -690,13 +752,16 @@ def collapse(iterable, base_type=None, levels=None):
         yield x
 
 
-def side_effect(func, iterable, chunk_size=None):
+def side_effect(func, iterable, chunk_size=None, before=None, after=None):
     """Invoke *func* on each item in *iterable* (or on each *chunk_size* group
     of items) before yielding the item.
 
-    `func` must be a function that takes a single argument.  Its return value
+    `func` must be a function that takes a single argument. Its return value
     will be discarded.
 
+    *before* and *after* are optional functions that take no arguments. They
+    will be executed before iteration starts and after it ends, respectively.
+
     `side_effect` can be used for logging, updating progress bars, or anything
     that is not functionally "pure."
 
@@ -717,16 +782,36 @@ def side_effect(func, iterable, chunk_size=None):
         >>> list(pair_sums)
         [1, 5, 9]
 
+    Writing to a file-like object:
+
+        >>> from io import StringIO
+        >>> from more_itertools import consume
+        >>> f = StringIO()
+        >>> func = lambda x: print(x, file=f)
+        >>> before = lambda: print(u'HEADER', file=f)
+        >>> after = f.close
+        >>> it = [u'a', u'b', u'c']
+        >>> consume(side_effect(func, it, before=before, after=after))
+        >>> f.closed
+        True
+
     """
-    if chunk_size is None:
-        for item in iterable:
-            func(item)
-            yield item
-    else:
-        for chunk in chunked(iterable, chunk_size):
-            func(chunk)
-            for item in chunk:
+    try:
+        if before is not None:
+            before()
+
+        if chunk_size is None:
+            for item in iterable:
+                func(item)
                 yield item
+        else:
+            for chunk in chunked(iterable, chunk_size):
+                func(chunk)
+                for item in chunk:
+                    yield item
+    finally:
+        if after is not None:
+            after()
 
 
 def sliced(seq, n):
@@ -741,8 +826,8 @@ def sliced(seq, n):
         >>> list(sliced((1, 2, 3, 4, 5, 6, 7, 8), 3))
         [(1, 2, 3), (4, 5, 6), (7, 8)]
 
-    This function will only work for sliceable objects. For non-sliceable
-    iterable, see ``chunked()``.
+    This function will only work for iterables that support slicing.
+    For non-sliceable iterables, see ``chunked()``.
 
     """
     return takewhile(bool, (seq[i: i + n] for i in count(0, n)))
@@ -760,7 +845,7 @@ def split_before(iterable, pred):
 
     """
     buf = []
-    for item in iter(iterable):
+    for item in iterable:
         if pred(item) and buf:
             yield buf
             buf = []
@@ -780,7 +865,7 @@ def split_after(iterable, pred):
 
     """
     buf = []
-    for item in iter(iterable):
+    for item in iterable:
         buf.append(item)
         if pred(item) and buf:
             yield buf
@@ -845,7 +930,9 @@ def distribute(n, iterable):
         >>> [list(c) for c in children]
         [[1], [2], [3], [], []]
 
-    This function uses ``itertools.tee``, and may require significant storage.
+    This function uses ``itertools.tee`` and may require significant storage.
+    If you need the order items in the smaller iterables to match the original
+    iterable, see ``divide()``.
 
     """
     if n < 1:
@@ -957,3 +1044,173 @@ def sort_together(iterables, key_list=(0,), reverse=False):
     return list(zip(*sorted(zip(*iterables),
                             key=itemgetter(*key_list),
                             reverse=reverse)))
+
+
+def divide(n, iterable):
+    """Divide the elements from *iterable* into *n* parts, maintaining
+    order.
+
+        >>> group_1, group_2 = divide(2, [1, 2, 3, 4, 5, 6])
+        >>> list(group_1)
+        [1, 2, 3]
+        >>> list(group_2)
+        [4, 5, 6]
+
+    If the length of the iterable is not evenly divisible by n, then the
+    length of the smaller iterables will not be identical::
+
+        >>> children = divide(3, [1, 2, 3, 4, 5, 6, 7])
+        >>> [list(c) for c in children]
+        [[1, 2, 3], [4, 5], [6, 7]]
+
+    If the length of the iterable is smaller than n, then the last returned
+    iterables will be empty:
+
+        >>> children = divide(5, [1, 2, 3])
+        >>> [list(c) for c in children]
+        [[1], [2], [3], [], []]
+
+    This function will exhaust the iterable before returning and may require
+    significant storage. If order is not important, see ``distribute()``,
+    which does not first pull the iterable into memory.
+
+    """
+    if n < 1:
+        raise ValueError('n must be at least 1')
+
+    seq = tuple(iterable)
+    q, r = divmod(len(seq), n)
+
+    ret = []
+    for i in range(n):
+        start = (i * q) + (i if i < r else r)
+        stop = ((i + 1) * q) + (i + 1 if i + 1 < r else r)
+        ret.append(iter(seq[start:stop]))
+
+    return ret
+
+
+def always_iterable(obj):
+    """
+    Given an object, always return an iterable.
+
+    If the object is not already iterable, return a tuple containing containing
+    the object::
+
+        >>> always_iterable(1)
+        (1,)
+
+    If the object is ``None``, return an empty iterable::
+
+        >>> always_iterable(None)
+        ()
+
+    Otherwise, return the object itself::
+
+        >>> always_iterable([1, 2, 3])
+        [1, 2, 3]
+
+    Strings (binary or unicode) are not considered to be iterable::
+
+        >>> always_iterable('foo')
+        ('foo',)
+
+    This function is useful in applications where a passed parameter may be
+    either a single item or a collection of items::
+
+        >>> def item_sum(param):
+        ...     total = 0
+        ...     for item in always_iterable(param):
+        ...         total += item
+        ...
+        ...     return total
+        >>> item_sum(10)
+        10
+        >>> item_sum([10, 20])
+        30
+
+    """
+    if obj is None:
+        return ()
+
+    string_like_types = (text_type, binary_type)
+    if isinstance(obj, string_like_types) or not hasattr(obj, '__iter__'):
+        return obj,
+
+    return obj
+
+
+def adjacent(predicate, iterable, distance=1):
+    """Return an iterable over ``(bool, item)`` tuples where the ``item`` is
+    drawn from *iterable* and the ``bool`` indicates whether
+    that item satisfies the *predicate* or is adjacent to an item that does.
+
+    For example, to find whether items are adjacent to a ``3``::
+
+        >>> list(adjacent(lambda x: x == 3, range(6)))
+        [(False, 0), (False, 1), (True, 2), (True, 3), (True, 4), (False, 5)]
+
+    Set *distance* to change what counts as adjacent. For example, To find
+    whether items are two places away from a ``3``:
+
+        >>> list(adjacent(lambda x: x == 3, range(6), distance=2))
+        [(False, 0), (True, 1), (True, 2), (True, 3), (True, 4), (True, 5)]
+
+    This is useful for contextualizing the results of a search function.
+    For example, a code comparison tool might want to identify lines that
+    have changed, but also surrounding lines to give the viewer of the diff
+    context.
+
+    The predicate function will only be called once for each item in the
+    iterable.
+
+    See also ``groupby_transform()``, which can be used with this function
+    to group ranges of items with the same ``bool`` value.
+
+    """
+    # Allow distance=0 mainly for testing that it reproduces results with map()
+    if distance < 0:
+        raise ValueError('distance must be at least 0')
+
+    i1, i2 = tee(iterable)
+    padding = [False] * distance
+    selected = chain(padding, map(predicate, i1), padding)
+    adjacent_to_selected = map(any, windowed(selected, 2 * distance + 1))
+    return zip(adjacent_to_selected, i2)
+
+
+def groupby_transform(iterable, keyfunc=None, valuefunc=None):
+    """Make an iterator that returns consecutive keys and groups from the
+    *iterable*. *keyfunc* is a function used to compute a grouping key for each
+    item. *valuefunc* is a function for transforming the items after grouping.
+
+    *keyfunc* and *valuefunc* default to identity functions if they are not
+    specified. When *valuefunc* is not specified, ``groupby_transform`` is the
+    same as ``itertools.groupby()``.
+
+    For example, to group a list of numbers by rounding down to the nearest 10,
+    and then transform them into strings:
+
+        >>> iterable = [0, 1, 12, 13, 23, 24]
+        >>> keyfunc = lambda x: 10 * (x // 10)
+        >>> valuefunc = lambda x: str(x)
+        >>> grouper = groupby_transform(iterable, keyfunc, valuefunc)
+        >>> [(k, list(g)) for k, g in grouper]
+        [(0, ['0', '1']), (10, ['12', '13']), (20, ['23', '24'])]
+
+    ``groupby_transform`` is useful when grouping elements of an iterable using
+    a separate iterable as the key. To do this,  ``zip()`` the iterables
+    and pass a *keyfunc* that extracts the first element and a *valuefunc*
+    that extracts the second element::
+
+        >>> from operator import itemgetter
+        >>> keys = [0, 0, 1, 1, 1, 2, 2, 2, 3]
+        >>> values = 'abcdefghi'
+        >>> iterable = zip(keys, values)
+        >>> grouper = groupby_transform(iterable, itemgetter(0), itemgetter(1))
+        >>> [(k, ''.join(g)) for k, g in grouper]
+        [(0, 'ab'), (1, 'cde'), (2, 'fgh'), (3, 'i')]
+
+    """
+    valuefunc = (lambda x: x) if valuefunc is None else valuefunc
+    return ((k, map(valuefunc, g)) for k, g in groupby(iterable, keyfunc))
